from types import MethodType


class RouteMeta:
    """
    路由元数据
    """

    def __init__(self,
                 handler: MethodType,
                 func_args,
                 route_id=None,
                 module=None,
                 name=None,
                 permission=True,
                 ajax=True,
                 referer=None,
                 kwargs=None):
        """

        :param handler: 路由处理函数对象
        :param func_args: 路由处理函数参数列表
        :param route_id: 路由ID，此ID由路由相关信息组合而成
        :param module: 装饰器上指定的 module 值
        :param name: 装饰器上指定的 name 值
        :param permission: 装饰器上指定的 name 值
        :param ajax: 装饰器上指定的 name 值
        :param referer: 装饰器上指定的 name 值
        :param kwargs: 装饰器上指定的其它参数
        """
        self._handler = handler
        self._func_args = func_args
        self._id = route_id
        self._module = module
        self._name = name
        self._permission = permission
        self._ajax = ajax
        self._referer = referer
        self._kwargs = {} if kwargs is None else kwargs

    @property
    def handler(self) -> MethodType:
        """
        路由处理函数对象
        :return:
        """
        return self._handler

    @property
    def func_args(self):
        """
        路由处理函数参数列表
        :return:
        :rtype: OrderedDict
        """
        return self._func_args

    @property
    def id(self) -> str:
        """
        路由ID，此ID由路由相关信息组合而成
        :return:
        """
        return self._id

    @property
    def module(self) -> str:
        """
        装饰器上指定的 module 值
        :return:
        """
        return self._module

    @property
    def name(self) -> str:
        """
        装饰器上指定的 name 值
        :return:
        """
        return self._name

    @property
    def permission(self) -> bool:
        """
        装饰器上指定的 permission 值
        :return:
        """
        return self._permission

    @property
    def ajax(self) -> bool:
        """
        装饰器上指定的 ajax 值
        :return:
        """
        return self._ajax

    @property
    def referer(self) -> str:
        """
        装饰器上指定的 referer 值
        :return:
        """
        return self._referer

    @property
    def kwargs(self) -> dict:
        """
        装饰器上指定的其它参数
        :return:
        :rtype: Dict
        """
        return self._kwargs



# coding: utf-8

# 此模块用于收集路由

import os
import re
from os import path

from django.conf import settings

# 全局类列表
GLOBAL_CLASSES = []


def register_globals(*global_classes):
    """
    注册全局类型
    :type global_classes: class
    :param global_classes:
    :return:
    """
    for cls in global_classes:
        GLOBAL_CLASSES.append(cls)


def _get_env(*args):
    env = {}

    # 加载全局配置
    for arg in GLOBAL_CLASSES:
        env[arg.__name__] = arg

    # 加载接口参数
    # 若存在相同的名称，则会被覆盖
    for arg in args:
        env[arg.__name__] = arg
    return env


def _get_route_map():
    from ..router import ROUTES_MAP
    if not ROUTES_MAP:
        raise Exception('[restful-dj] Routes map is empty, did you forgot to call `restful_dj.map_routes(*routes_map)`')

    return ROUTES_MAP.items()


def collect(*environments):
    """
    执行收集操作
    :return: 所有路由的集合
    """
    # 为 route 提供的执行环境
    # 读取在 settings.py 中配置的环境
    route_env = _get_env(fake_route, *environments)
    project_root = settings.BASE_DIR

    # 所有路由的集合
    routes = []

    for (http_prefix, pkg_prefix) in _get_route_map():
        route_root = path.abspath(path.join(project_root, pkg_prefix.replace('.', path.sep)))

        # 遍历目录，找出所有的 .py 文件
        for (dir_name, dirs, files) in os.walk(route_root):
            for file in files:
                # 不是 .py 文件，忽略
                if not file.endswith('.py'):
                    continue

                fullname = path.abspath(path.join(dir_name, file))
                # 解析文件
                get_route_defines(route_root, fullname, http_prefix, pkg_prefix, routes, route_env)

            for sub_dir_name in dirs:
                pkg_file = path.abspath(path.join(sub_dir_name, '__init__.py'))
                if path.exists(pkg_file) and path.isfile(pkg_file):
                    get_route_defines(route_root, pkg_file, http_prefix, pkg_prefix, routes, route_env)

    return routes


def get_route_defines(route_root, fullname, http_prefix, pkg_prefix, routes, route_env):
    for define in resolve_file(route_root, fullname, http_prefix, pkg_prefix, route_env):
        # 返回 None 表示没有找到定义
        if define is None:
            continue
        routes.append(define)


def resolve_file(route_define, fullname, http_prefix, pkg_prefix, route_env: dict):
    """
    解析文件
    :param route_env:
    :param pkg_prefix:
    :param http_prefix: http 请求前缀
    :param route_define: 路由文件的根路径
    :param fullname: 文件的完整路径
    :return: 没有路由时返回 None
    """
    with open(fullname, encoding='utf-8') as python_fp:
        lines = python_fp.readlines()
        python_fp.close()
    source = ''.join(lines)

    # 解析路由的定义
    # defines = re.compile(r'^@(route\(.+?\)).*(\n.*?)+^def (.+?)\(', re.M).findall(source)
    defines = re.compile(r'^@(route\(.+?\))(.*?)^def (.+?)\(', re.M | re.S).findall(source)
    # 没有找到定义，返回 None
    if len(defines) == 0:
        yield None

    # router_str 是在函数上声明的装饰器定义
    # other_lines 是装饰器与函数声明间的其它代码
    # func 是函数的名称
    for (router_str, other_lines, func) in defines:
        # 解析出请求的方法(method)与请求的指定函数名称
        method, name = resolve_func(func)
        # 利用 eval 解析出路由的定义（在这个文件中定义了与装饰器相同的函数，以便于读取装饰器的参数）
        router_str = router_str.replace('route', 'fake_route')
        define = eval(router_str, route_env)

        # 构造http请求的地址
        # -3 是为了干掉最后的 .py 字样
        pkg = re.sub(r'[/\\]', '.', path.relpath(fullname, route_define))[0:-3]

        # 当是包时，移除 __init__ 部分
        if path.basename(fullname) == '__init__.py':
            http_path = '%s.%s' % (http_prefix, pkg[0:-len('__init__')])
        else:
            http_path = '%s.%s' % (http_prefix, pkg)

        # 当路由文件为根目录下的 __init__.py 时，没有可访问的文件名
        # 此时会出现得到的路由为  xxx. 的情况
        # 所以在此移除末尾的 . 符号
        http_path = http_path.rstrip('.')

        # 如果指定了名称，就追加到地址后
        if name is not None:
            http_path += '/' + name

        pkg = '%s.%s' % (pkg_prefix, pkg)
        # 唯一标识
        define['id'] = '%s_%s' % (pkg.replace('_', '__').replace('.', '_'), func)
        # 路由所在包名称
        define['pkg'] = pkg
        # 路由所在文件的完整路径
        define['file'] = fullname
        # 路由请求的处理函数
        define['handler'] = func
        # 路由的请求方法
        define['method'] = method
        # 路由的请求路径
        define['path'] = http_path

        yield define


def fake_route(module=None, name=None, permission=True, ajax=True, referer=None, **kwargs):
    """
    此函数用于帮助读取装饰器的参数
    :param module:
    :param name:
    :param permission:
    :param ajax:
    :param referer:
    :param kwargs:
    :return:
    """

    return {
        'module': module,
        'name': name,
        'permission': permission,
        'ajax': ajax,
        'referer': referer,
        'kwargs': kwargs
    }


def resolve_func(func: str):
    """
    从处理函数中解析出路由的 method 与指定名称
    :param func:
    :return:
    """
    method, lodash, name = re.match(r'([a-z]+)(_(.+))?', func).groups()
    return method, name


# 生成：注册路由的代码 -- 模板
# 注意生成的代码中的缩进，使用的是空格
_REGISTER_STMT = "    # {module}-{name}\n    ['{method}', '{path}', {handler}]"
_CODE_TPL = """# -*- coding={encoding} -*-

# IMPORT ROUTES BEGIN
{imports}
# IMPORT ROUTES END


# REGISTER ROUTES BEGIN
routes = [
{routes}
]
# REGISTER ROUTES END
"""


def persist(filename: str = '', encoding='utf8'):
    """
    将路由持久化
    :param filename:
    :param encoding:
    :return: 持久化的 python 代码
    :rtype: str or None
    """
    imports = []
    routes = []

    print('[restful-dj] Generating restful map file with encoding %s' % encoding)
    for route in collect():
        # imports.append('from %s import %s as %s' % (route['pkg'], route['handler'], route['id']))
        imports.append('from %s import %s as %s' % (route['pkg'], route['handler'], route['id']))
        routes.append(_REGISTER_STMT.format(
            module=route['module'],
            name=route['name'],
            method=route['method'].upper(),
            path=route['path'],
            handler=route['id']
        ))

    content = _CODE_TPL.format(encoding=encoding, imports='\n'.join(imports), routes=',\n'.join(routes))

    print('[restful-dj] Generate restful map file complete')
    if not filename:
        return content

    print('[restful-dj] Persisting into file %s' % filename)
    with open(filename, mode='wt', encoding=encoding) as fp:
        fp.write(content)
        fp.close()
    print('[restful-dj] Persist into file complete')



from django.http import HttpResponse, HttpRequest

from .meta import RouteMeta

# 注册的路由中间件列表
MIDDLEWARE_INSTANCE_LIST = []


def register_middlewares(*middlewares):
    """
    注册中间件，注册的中间件将按顺序执行
    :param middlewares:
    :return:
    """
    for middleware_cls in middlewares:
        MIDDLEWARE_INSTANCE_LIST.append(middleware_cls())


# noinspection PyMethodMayBeStatic,PyUnusedLocal
class MiddlewareBase:
    """
    路由中间件基类
    """

    def process_request(self, request: HttpRequest, meta: RouteMeta, **kwargs):
        """
        对 request 对象进行预处理。一般用于请求的数据的解码，此时路由组件尚水进行请求数据的解析(B,P,G 尚不可用)
        :param request:
        :param meta:
        :return: 返回 HttpResponse 以终止请求，返回 False 以停止执行后续的中间件(表示访问未授权)，返回 None 或不返回任何值继续执行后续中间件
        """
        pass

    def process_invoke(self, request: HttpRequest, meta: RouteMeta, **kwargs):
        """
        在路由函数调用前，对其参数等进行处理，此时路由组件已经完成了请求数据的解析(B,P,G 已可用)
        此时可以对解析后的参数进行变更
        :param request:
        :param meta:
        :return: 返回 HttpResponse 以终止请求，返回 False 以停止执行后续的中间件(表示访问未授权)，返回 None 或不返回任何值继续执行后续中间件
        """
        pass

    def process_return(self, request: HttpRequest, meta: RouteMeta, **kwargs):
        """
        在路由函数调用后，对其返回值进行处理
        :param request:
        :param meta:
        :param kwargs: 始终会有一个 'data' 的项，表示路由返回的原始数据
        :return: 返回 HttpResponse 以终止执行，否则返回新的 数据
        """
        assert 'data' in kwargs
        return kwargs['data']

    def process_response(self, request: HttpRequest, meta: RouteMeta, **kwargs) -> HttpResponse:
        """
        对 response 数据进行预处理。一般用于响应的数据的编码
        :param request:
        :param meta:
        :param kwargs: 始终会有一个 'response' 的项，表示路由返回的原始 HttpResponse
        :return: 无论何种情况，应该始终返回一个  HttpResponse
        :rtype: HttpResponse
        """
        assert 'response' in kwargs
        return kwargs['response']


class MiddlewareManager:
    """
    路由中间件管理器
    """

    def __init__(self, request: HttpRequest, meta: RouteMeta):
        # HTTP请求对象
        self.request = request
        # 元数据信息
        self.meta = meta

    def begin(self):
        for middleware in MIDDLEWARE_INSTANCE_LIST:
            if not hasattr(middleware, 'process_request'):
                continue
            result = middleware.process_request(self.request, self.meta)
            if isinstance(result, HttpResponse):
                return result
            # 返回 False 以阻止后续中间件执行
            if result is False:
                return

    def before_invoke(self):
        """
        在路由函数调用前，对其参数等进行处理
        :return:
        """
        for middleware in MIDDLEWARE_INSTANCE_LIST:
            if not hasattr(middleware, 'process_invoke'):
                continue
            result = middleware.process_invoke(self.request, self.meta)
            if isinstance(result, HttpResponse):
                return result
            # 返回 False 以阻止后续中间件执行
            if result is False:
                return

    def process_return(self, data):
        """
        在路由函数调用后，对其返回值进行处理
        :param data:
        :return:
        """
        for middleware in MIDDLEWARE_INSTANCE_LIST:
            if not hasattr(middleware, 'process_return'):
                continue
            result = middleware.process_return(self.request, self.meta, data=data)

            # 返回 HttpResponse 终止
            if result is HttpResponse:
                return result

            # 使用原数据
            data = result

        return data

    def end(self, response):
        """
        在响应前，对响应的数据进行处理
        :param response:
        :return:
        """
        # 对 response 进行处理
        for middleware in MIDDLEWARE_INSTANCE_LIST:
            if not hasattr(middleware, 'process_response'):
                continue
            response = middleware.process_response(self.request, self.meta, response=response)

        return response



from django.conf import settings

CUSTOMIZE_LOGGER = None


def set_logger(logger):
    """
    设置自定义的 logger
    :param logger:
    :return:
    """
    global CUSTOMIZE_LOGGER
    CUSTOMIZE_LOGGER = logger


def log(level, message, e=None):
    if CUSTOMIZE_LOGGER is None:
        print('[%s] %s' % (level, message))
    else:
        # noinspection PyUnresolvedReferences
        CUSTOMIZE_LOGGER.log(level, message, e)


def debug(message):
    log('debug', message)


def success(message):
    log('success', message)


def info(message):
    log('info', message)


def warning(message):
    log('warning', message)


def error(message, e=None, _raise=True):
    temp = message if e is None else '%s\n\t%s' % (message, repr(e))

    # 非开发模式时，始终不会输出堆栈信息
    if not settings.DEBUG:
        log('ERROR', temp, e)
        return

    # print('\033[1;31;47m {0} \033[0m'.format(temp))
    # if e is not None:
    #     print(repr(e.__traceback__.tb_frame))

    # 不需要抛出异常
    if not _raise:
        log('ERROR', temp, e)
        return

    # 抛出新的异常
    if e is None:
        raise Exception(message)

    # 修改异常消息
    new_msg = '%s\n\t%s' % (message, e.args[0]) if len(e.args) > 0 else message
    e.args = (new_msg,)
    raise e



import inspect
import re
from collections import OrderedDict

from django.http import HttpRequest

from . import logger


def _get_parameter_alias(match):
    """

    :param match:
    :type match: Match
    :return:
    """
    ch = match.group('ch')
    return '' if ch is None else ch.upper()


class ArgumentSpecification:
    """
    函数参数声明
    """

    def __init__(self, name: str, index: int):
        """

        :param name: 参数名称
        :param index: 参数在参数位置中的位置
        """
        self.name = name
        self.index = index
        # 是否是可变参数
        self.is_variable = False
        # 是否有类型声明
        self.has_annotation = False
        # 是否有默认值
        self.has_default = False
        # 类型声明
        self.annotation = None
        # 默认值
        self.default = None
        # 注释
        self.comment = None
        # 别名，当路由处理函数中声明的是 abc_def 时，自动处理为 abcDef
        # 同时会移除所有的 _ 符号
        self.alias = re.sub('_+(?P<ch>.?)', _get_parameter_alias, name)
        # 如果与原名称相同，那么就设置为 None 表示无别名
        if self.alias == name:
            self.alias = None

    @property
    def annotation_name(self):
        return self.annotation.__name__ if self.has_annotation else 'any'

    def __str__(self):
        arg_type = self.annotation.__name__ if self.has_annotation else 'any'

        name = self.name

        name = name if self.alias is None else '%s/%s' % (name, self.alias)

        if self.has_default:
            default_value = "'%s'" % self.default if isinstance(self.default, str) else self.default
            return '%s: %s=%s' % (name, arg_type, default_value)

        return '%s: %s' % (name, arg_type)


def get_func_docs(func):
    docs = {}

    # :type str
    doc_str = func.__doc__
    if doc_str is None:
        return docs

    temp = doc_str.splitlines(False)

    for row in temp:
        if not row:
            continue
        match = re.match(r'\s*:param\s+(?P<name>[\S]+):(?P<comment>.*)$', row)
        if not match:
            continue
        docs[match.group('name')] = match.group('comment')

    return docs


def get_func_args(func):
    """
    获取函数的参数列表（带参数类型）
    :param func:
    :return:
    """
    signature = inspect.signature(func)
    parameters = signature.parameters
    _empty = signature.empty

    documatation = get_func_docs(func)

    args = OrderedDict()
    index = 0
    for p in parameters.keys():
        parameter = parameters.get(p)
        spec = ArgumentSpecification(p, index)
        spec.is_variable = parameter.kind == parameter.VAR_KEYWORD

        if p in documatation:
            spec.comment = documatation[p]

        index += 1
        # 类型
        annotation = parameter.annotation

        # 无效的类型声明
        if not inspect.isclass(annotation):
            source_lines = inspect.getsourcelines(func)
            line = source_lines[1]
            row = None
            for row in source_lines[0]:
                if parameter.name in row:
                    break
                line += 1
            msg = 'File "%s", line %d, in %s\n\t' % (
                inspect.getmodule(func).__file__,
                line,
                func.__name__
            )
            msg += 'Invalid type declaration for argument "%s":\n\t\t%s'
            # abort the execution
            logger.error(msg % (parameter.name, row.strip()))

        default = parameter.default

        if default != _empty:
            spec.default = default
            spec.has_default = True

        # 有默认值时，若未指定类型，则使用默认值的类型
        if annotation == _empty:
            if default is not None and default != _empty:
                spec.annotation = type(default)
                spec.has_annotation = True
            elif p == 'request':
                # 以下情况将设置为 HttpRequest 对象
                # 1. 当参数名称是 request 并且未指定类型
                # 2. 当参数类型是 HttpRequest 时 (不论参数名称，包括 request)
                # 但是，参数名称是 request 但其类型不是 HttpRequest ，就会被当作一般参数处理
                spec.annotation = HttpRequest
                spec.has_annotation = True
        else:
            spec.annotation = annotation
            spec.has_annotation = True

        args[p] = spec

    return args


def load_module(module_name: str):
    """
    加载模块
    :param module_name:
    :return:
    """
    # noinspection PyTypeChecker
    return __import__(module_name, fromlist=True)


def get_func_info(func):
    source_lines = inspect.getsourcelines(func)
    line = source_lines[1]
    return 'File "%s", line %d, in %s' % (
        inspect.getmodule(func).__file__,
        line,
        func.__name__
    )



import json
from collections import OrderedDict
from functools import wraps

from django.http import HttpResponse, JsonResponse, HttpResponseBadRequest, HttpRequest

from .meta import RouteMeta
from .middleware import MiddlewareManager
from .util import logger
from .util.utils import ArgumentSpecification
from .util.utils import get_func_info


def route(module=None, name=None, permission=True, ajax=True, referer=None, **kwargs):
    """
    用于控制路由的访问权限，路由均需要添加此装饰器，若未添加，则不可访问
    :param module: str 路由所属模块，一般在查询权限时会用到
    :param name: str 路由名称，一般在查询权限时会用到
    :param permission: bool 访问此地址是否需要检查用户权限,由数据库实际值控制,在这里只起到做入库作用
    :param ajax: bool 是否仅允许ajax请求,由数据库实际值控制,在这里只起到做入库作用
    :param referer: list|str 允许的来源页,由数据库实际值控制,在这里只起到做入库作用
    用法：
    @route('用户管理', '编辑用户', permission=True)
    def get(req):
        pass
    """

    def invoke_route(func):
        @wraps(func)
        def caller(*args):
            # 参数长度不为 2 时，认为是用户调用
            if len(args) != 2:
                return func(*args)

            # Http 请求对象
            # :type HttpRequest
            request = args[0]
            # 函数声明时定义的参数列表
            # :type OrderedDict
            func_args = args[1]

            # 如果传入的参数第一个不是 request，第二个不是 OrderedDict，
            # 那么就认为是用户调用，而不是路由调用
            # 此时直接将原参数传给 func 进行调用
            if not isinstance(request, HttpRequest) or not isinstance(func_args, OrderedDict):
                return func(*args)

            meta = RouteMeta(
                func,
                func_args,
                route_id='%s_%s' % (func.__module__.replace('_', '__').replace('.', '_'), func.__name__),
                module=module,
                name=name,
                permission=permission,
                ajax=ajax,
                referer=referer,
                kwargs=kwargs,
            )

            return _invoke_with_route(request, meta)

        return caller

    return invoke_route


def _invoke_with_route(request: HttpRequest, meta: RouteMeta):
    mgr = MiddlewareManager(
        request,
        meta
    )

    func_args = meta.func_args
    func = meta.handler

    # 调用中间件，以处理请求
    result = mgr.begin()

    # 返回了 HttpResponse，直接返回此对象
    if isinstance(result, HttpResponse):
        return mgr.end(result)

    # 返回了 False，表示未授权访问
    if result is False:
        return mgr.end(HttpResponseUnauthorized())

    # 处理请求中的json参数
    # 处理后可能会在 request 上添加一个 json 的项，此项存放着json格式的 body 内容
    # noinspection PyTypeChecker
    _process_json_params(request)

    result = mgr.before_invoke()

    # 返回了 False，表示未授权访问
    if result is False:
        return mgr.end(HttpResponseUnauthorized())

    # 返回了 HttpResponse ， 直接返回此对象
    if isinstance(result, HttpResponse):
        return mgr.end(result)

    # 调用路由处理函数
    arg_len = len(func_args)
    if arg_len == 0:
        return mgr.end(_wrap_http_response(mgr, func()))

    # 有参数，自动从 queryString, POST 或 json 中获取
    # 匹配参数

    actual_args = _get_actual_args(request, func, func_args)

    if isinstance(actual_args, HttpResponse):
        return mgr.end(actual_args)

    result = func(**actual_args)

    return mgr.end(_wrap_http_response(mgr, result))


def _process_json_params(request):
    """
    参数处理
    :return:
    """
    request.B = {}
    request.G = {}
    request.P = {}

    if request.content_type != 'application/json':
        request.G = request.GET.dict()
        request.P = request.POST.dict()
        return

    # 如果请求是json类型，就先处理一下

    body = request.body

    if body == '' or body is None:
        return

    try:
        if isinstance(body, str):
            request.B = json.loads(body)
        elif isinstance(body, bytes):
            request.B = json.loads(body.decode())
        elif isinstance(body, (dict, list)):
            request.B = body
    except Exception as e:
        logger.warning('Deserialize request body fail: %s' % str(e))


def _get_parameter_str(args: OrderedDict):
    return '\n\t\t'.join([str(args[arg]) for arg in args])


def _get_value(data: dict, name: str, arg_spec: ArgumentSpecification, backup: dict = None):
    """

    :param data:
    :param name:
    :param arg_spec:
    :return: True 表示使用默认值 False 表示未使用默认值 None 表示无值
    """
    if name in data:
        return False, data[name]

    alias = arg_spec.alias

    if alias is not None and alias in data:
        return False, data[alias]

    if backup is not None:
        if name in backup:
            return False, backup[name]

        if alias is not None and alias in backup:
            return False, backup[alias]

    # 使用默认值
    if arg_spec.has_default:
        return True, arg_spec.default

    # 缺少无默认值的参数
    return None, None


def _get_actual_args(request: HttpRequest, func, args: OrderedDict) -> dict or HttpResponse:
    method = request.method.lower()
    actual_args = {}

    # 已使用的参数名称，用于后期填充可变参数时作排除用
    used_args = []
    # 是否声明了可变参数
    has_variable_args = False

    # noinspection PyUnresolvedReferences
    arg_source = request.G if method in ['delete', 'get'] else request.P

    for arg_name in args.keys():
        arg_spec = args.get(arg_name)

        # 如果是可变参数：如: **kwargs
        # 设置标记，以在后面进行填充
        if arg_spec.is_variable:
            has_variable_args = True
            continue

        # 以及情况将传入 HttpRequest 对象
        # 1. 当参数名称是 request 并且未指定类型
        # 2. 当参数类型是 HttpRequest 时 (不论参数名称，包括 request)
        # 但是，参数名称是 request 但其类型不是 HttpRequest ，就会被当作一般参数处理
        if arg_name == 'request' and not arg_spec.has_annotation:
            actual_args[arg_name] = request
            continue
        if arg_spec.annotation == HttpRequest:
            actual_args[arg_name] = request
            continue

        # noinspection PyUnresolvedReferences
        use_default, arg_value = _get_value(arg_source, arg_name, arg_spec, request.B)

        # 未找到参数
        if use_default is None:
            msg = '%s\n\tMissing required argument "%s":\n\t\t%s' % (
                get_func_info(func),
                arg_name,
                _get_parameter_str(args)
            )
            logger.warning(msg)
            return HttpResponseBadRequest(msg)

        # 使用默认值
        if use_default is True:
            actual_args[arg_name] = arg_value
            used_args.append(arg_name)
            continue

        # 未指定类型
        if not arg_spec.has_annotation:
            actual_args[arg_name] = arg_value
            used_args.append(arg_name)
            continue

        # 当值为 None 时，不作数据类型校验
        if arg_value is None:
            actual_args[arg_name] = arg_value
            used_args.append(arg_name)
            continue

        # 检查类型是否一致 #

        # 类型一致，直接使用
        if isinstance(arg_value, arg_spec.annotation):
            actual_args[arg_name] = arg_value
            used_args.append(arg_name)
            continue

        # 类型不一致，尝试转换类型

        # 当声明的参数类型是布尔类型时，收到的值可能是一个字符串（其值为 true 、 false）
        if arg_spec.annotation is bool and isinstance(arg_value, str):
            if arg_value == 'true':
                actual_args[arg_name] = True
                used_args.append(arg_name)
            elif arg_value == 'false':
                actual_args[arg_name] = False
                used_args.append(arg_name)
            else:
                logger.error('Value for "%s!%s" may be incorrect (a boolean value expected: true/false): %s' % (
                    func.__name__, arg_name, arg_value))
            continue

        # 转换失败时，会抛出异常
        # noinspection PyBroadException
        try:
            # 当 arg_value 是字符串，arg_spec的类型是对象时，尝试解析成 json
            if arg_spec.annotation in (dict, list) and isinstance(arg_value, str):
                # noinspection PyBroadException
                try:
                    arg_value = json.loads(arg_value)
                except Exception:
                    # 此处的异常直接忽略即可
                    logger.warning('Value for "%s!%s" may be incorrect: %s' % (func.__name__, arg_name, arg_value))

                # 类型一致，直接使用
                if isinstance(arg_value, arg_spec.annotation):
                    actual_args[arg_name] = arg_value
                    used_args.append(arg_name)
                    continue
            actual_args[arg_name] = arg_spec.annotation(arg_value)
            used_args.append(arg_name)
        except Exception:
            msg = 'Argument type of "%s" mismatch, expect type "%s" but got "%s", signature: (%s)' \
                  % (arg_name, arg_spec.annotation.__name__, type(arg_value).__name__, _get_parameter_str(args))
            logger.warning(msg)
            return HttpResponseBadRequest(msg)

    if not has_variable_args:
        return actual_args

    # 填充可变参数
    variable_args = {}
    for item in arg_source:
        if item in used_args:
            continue
        variable_args[item] = arg_source[item]

    # noinspection PyUnresolvedReferences
    for item in request.B:
        if item in used_args:
            continue
        # noinspection PyUnresolvedReferences
        variable_args[item] = request.B[item]

    actual_args.update(variable_args)

    return actual_args


def _wrap_http_response(mgr, data):
    """
    将数据包装成 HttpResponse 返回
    :param data:
    :return:
    """

    # 处理返回函数
    data = mgr.process_return(data)

    if data is None:
        return HttpResponse()

    if isinstance(data, HttpResponse):
        return data

    if isinstance(data, bool):
        return HttpResponse('true' if bool else 'false')

    if isinstance(data, (dict, list, set, tuple)):
        return JsonResponse(data, safe=False)

    if isinstance(data, str):
        return HttpResponse(data.encode())

    if isinstance(data, bytes):
        return HttpResponse(data)

    return HttpResponse(str(data).encode())


class HttpResponseUnauthorized(HttpResponse):
    status_code = 401



import inspect
import os
from types import MethodType

from django import shortcuts
from django.conf import settings
from django.http import HttpResponseNotFound, HttpResponseServerError, HttpRequest, HttpResponse

from .util import collector, utils
from .util import logger
from .util.utils import load_module

# 包名称
NAME = 'restful_dj'

# 函数缓存，减少 inspect 反射调用次数
ENTRY_CACHE = {}

_BEFORE_DISPATCH_HANDLER = None

# 线上模式时，使用固定路由
PRODUCTION_ROUTES = {}

# 开发模式的模块缓存，用于API列表
MODULES_CACHE = None

# 路由映射表，其键为请求的路径，其值为映射的目录
ROUTES_MAP = {}


def map_routes(routes_map: dict):
    """
    注册路由映射表
    :param routes_map:
    :return:
    """
    for path in routes_map:
        ROUTES_MAP[path] = routes_map[path]


def register_routes(routes: list):
    """
    手动注册路由列表
    :param routes: 其每一项都应该是一个 list, 元素依次为 method: str, path: str, handler: MethodType
    :return:
    """
    for _route in routes:
        register(*_route)


def register(method: str, path: str, handler: MethodType):
    """
    手动注册路由
    :param path:
    :param method:
    :param handler:
    :return:
    """
    rid = '%s#%s' % (path, method.lower())
    if rid in PRODUCTION_ROUTES:
        logger.warning('[restful-dj] %s %s exists' % (method, path))

    args = utils.get_func_args(handler)
    PRODUCTION_ROUTES[rid] = {
        'func': handler,
        'args': args
    }


def set_before_dispatch_handler(handler):
    """
    设置请求分发前的处理函数
    :param handler:
    :return:
    """
    global _BEFORE_DISPATCH_HANDLER
    _BEFORE_DISPATCH_HANDLER = handler


# noinspection PyUnusedLocal
def redirect(request, entry, name=''):
    """
    当请求的URL使用了 / 结尾时，重定向到不使用 / 结尾的地址
    :param request:
    :param entry:
    :param name:
    :return:
    """
    return shortcuts.redirect(request.path.rstrip('/'))


def render_list(request):
    if not settings.DEBUG:
        return HttpResponseNotFound()

    global MODULES_CACHE

    prefix = '%s://%s%s' % (request.scheme, request.META.get('HTTP_HOST'), request.path)

    if MODULES_CACHE is None:
        routes = collector.collect()

        modules = {}

        for route in routes:
            module = route['module']

            p = route['path']
            suffix = ''
            temp = p.split('/')
            entry = temp[0]
            if len(temp) == 2:
                suffix = temp[1]

            router = Router(request, route['method'], entry, suffix)
            router.check()
            define = router.get_func_define()
            if define and len(define['args']) > 0:
                route['args'] = [define['args'][arg] for arg in define['args']]
            else:
                route['args'] = None

            if module in modules:
                modules[module].append(route)
            else:
                modules[module] = [route]
        MODULES_CACHE = modules

    return shortcuts.render(request, 'restful_dj_api_list_template.html', {
        'modules': MODULES_CACHE,
        'prefix': prefix
    })


def dispatch(request, entry, name=''):
    """
    REST-ful 路由分发入口
    :param request: 请求
    :param entry: 入口文件，包名使用 . 符号分隔
    :param name='' 指定的函数名称
    :return:
    """
    if _BEFORE_DISPATCH_HANDLER is not None:
        # noinspection PyCallingNonCallable
        entry, name = _BEFORE_DISPATCH_HANDLER(request, entry, name)

    if not settings.DEBUG:
        return _route_for_production(request, entry, name)

    router = Router(request, request.method, entry, name)
    check_result = router.check()
    if isinstance(check_result, HttpResponse):
        return check_result
    return router.route()


def _route_for_production(request, entry, name):
    method = request.method.lower()
    # noinspection PyBroadException
    try:
        if name == '':
            route = PRODUCTION_ROUTES['%s#%s' % (entry, method)]
        else:
            route = PRODUCTION_ROUTES['%s/%s#%s' % (entry, name, method)]
    except Exception:
        return HttpResponseNotFound()

    return _invoke_handler(request, route['func'], route['args'])


def _invoke_handler(request, func, args):
    try:
        return func(request, args)
    except Exception as e:
        message = '[restful-dj]\n\t%s' % utils.get_func_info(func)
        logger.error(message, e)
        return HttpResponseServerError('%s: %s' % (message, str(e)))


class Router:
    def __init__(self, request: HttpRequest, method: str, entry: str, name: str):
        self.request = request
        self.entry = entry
        method = method.lower()
        self.method = method

        # 如果指定了名称，那么就加上
        # 如：name = 'detail'
        #   func_name = get_detail
        if name:
            func_name = '%s_%s' % (method, name.lower())
        else:
            func_name = method
        self.func_name = func_name

        # 处理映射
        # 对应的模块(文件路径）
        self.module_name = self.get_route_map(entry)

        self.fullname = ''

    def check(self):
        module_name = self.module_name

        if module_name is None:
            logger.warning('Cannot find route map in RESTFUL_DJ.routes: %s' % self.entry)
            return HttpResponseNotFound()

        # 如果 module_name 是目录，那么就查找 __init__.py 是否存在
        abs_path = os.path.join(settings.BASE_DIR, module_name.replace('.', os.path.sep))
        if os.path.isdir(abs_path):
            logger.info('Entry "%s" is package, auto load module "__init__.py"' % module_name)
            module_name = '%s.%s' % (module_name, '__init__')
        elif not os.path.exists('%s.py' % abs_path):
            return HttpResponseNotFound()

        self.module_name = module_name

        # 完全限定名称
        self.fullname = '%s.%s' % (module_name, self.func_name)

    def route(self):
        try:
            func_define = self.get_func_define()
        except Exception as e:
            message = 'Load entry "%s" failed' % self.module_name
            logger.error(message, e)
            return HttpResponseNotFound()

        # 如果 func_define 为 False ，那就表示此函数不存在
        if func_define is False:
            message = 'Route "%s.%s" not found' % (self.module_name, self.func_name)
            logger.info(message)
            return HttpResponseNotFound()

        if func_define is HttpResponse:
            return func_define

        return _invoke_handler(self.request, func_define['func'], func_define['args'])

    def get_func_define(self):
        fullname = self.fullname
        func_name = self.func_name
        module_name = self.module_name

        # 缓存中有这个函数
        if fullname in ENTRY_CACHE.keys():
            return ENTRY_CACHE[fullname]

        # 缓存中没有这个函数，去模块中查找
        # ---------------

        try:
            # 如果不加上fromlist=True,只会导入目录
            # noinspection PyTypeChecker
            # __import__ 自带缓存
            entry_define = load_module(module_name)
        except Exception as e:
            message = 'Load module "%s" failed' % module_name
            logger.error(message, e)
            return HttpResponseNotFound()

        # 模块中也没有这个函数
        if not hasattr(entry_define, func_name):
            # 函数不存在，更新缓存
            ENTRY_CACHE[func_name] = False
            return False

        # 模块中有这个函数
        # 通过反射从模块加载函数
        func = getattr(entry_define, func_name)
        if not self.is_valid_route(func):
            msg = '%s\n\tDecorator "@route" not found on function "%s", did you forgot it ?' % (
                utils.get_func_info(func),
                fullname
            )
            logger.warning(msg)
            # 没有配置装饰器@route，则认为函数不可访问，更新缓存
            ENTRY_CACHE[func_name] = False
            return False

        ENTRY_CACHE[fullname] = {
            'func': func,
            # 该函数的参数列表
            # 'name': {
            #     'annotation': '类型', 当未指定类型时，无此项
            #     'default': '默认值'，当未指定默认值时，无此项
            # }
            'args': utils.get_func_args(func)
        }

        return ENTRY_CACHE[fullname]

    @staticmethod
    def is_valid_route(func):
        source = inspect.getsource(func)
        lines = source.split('\n')
        for line in lines:
            if line.startswith('def '):
                # 已经查找到了函数定义部分了，说明没有找到
                return False

            if line.startswith('@route('):
                # 是 @route 装饰器行
                return True
        return False

    @staticmethod
    def get_route_map(route_path):
        # 命中
        hit_route = None
        for root_path in ROUTES_MAP:
            if route_path.startswith(root_path):
                hit_route = root_path, ROUTES_MAP[root_path]
                break

        if hit_route is None:
            return None

        # 将请求路径替换为指定的映射路径
        return ('%s%s' % (hit_route[1], route_path[len(hit_route[0]):])).strip('.')



import setuptools

VERSION = '2.0.0-beta1'

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setuptools.setup(
    name='restful-dj',
    version=VERSION,
    packages=setuptools.find_packages(),
    include_package_data=True,
    url='http://github.com/hyjiacan/restful-dj',
    license='MIT',
    author='hyjiacan',
    author_email='hyjiacan@163.com',
    description='restful(and auto route) support for Django2/3',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)



from django.conf import settings
from django.urls import path

from . import router
from .decorator import route
from .meta import RouteMeta
from .middleware import register_middlewares
from .router import set_before_dispatch_handler, register_routes, map_routes
from .util.collector import collect, persist, register_globals
from .util.logger import set_logger

dispatch = (
    [
        path('', router.render_list),
        path('<str:entry>', router.dispatch),
        path('<str:entry>/', router.redirect),
        path('<str:entry>/<str:name>', router.dispatch),
        path('<str:entry>/<str:name>/', router.redirect)
    ],
    router.NAME,
    router.NAME
)

__all__ = [
    'collect',
    'persist',
    'route',
    'RouteMeta',
    'set_before_dispatch_handler',
    'set_logger',
    'map_routes',
    'register_globals',
    'register_routes',
    'register_middlewares',
    'dispatch'
]

if settings.DEBUG:
    settings.INSTALLED_APPS.append('restful_dj')
